/*
 * PROJECT:         ReactOS Win32K
 * LICENSE:         GPL - See COPYING in the top level directory
 * FILE:            subsystems/win32/win32k/wine/winesup.c
 * PURPOSE:         Wine supporting functions
 * PROGRAMMERS:     Aleksey Bragin (aleksey@reactos.org)
 */

/* INCLUDES ******************************************************************/

#include <win32k.h>

#include "object.h"
#include "user.h"

#define NDEBUG
#include <debug.h>

timeout_t current_time = 0ULL;
int debug_level = 0;

unsigned int global_error;

typedef struct _SHELL_HOOK_WINDOW
{
  struct list ListEntry;
  HWND hWnd;
} SHELL_HOOK_WINDOW, *PSHELL_HOOK_WINDOW;

static struct list global_shell_hooks = LIST_INIT(global_shell_hooks);

struct desktop *get_desktop_obj( PPROCESSINFO process, obj_handle_t handle, unsigned int access );

/* PRIVATE FUNCTIONS *********************************************************/

void set_error( unsigned int err )
{
    global_error = err;
}

unsigned int get_error(void)
{
    return global_error;
}

const SID *token_get_user( void *token )
{
    UNIMPLEMENTED;
    return NULL;
}

/* default map_access() routine for objects that behave like an fd */
unsigned int default_fd_map_access( struct object *obj, unsigned int access )
{
    if (access & GENERIC_READ)    access |= FILE_GENERIC_READ;
    if (access & GENERIC_WRITE)   access |= FILE_GENERIC_WRITE;
    if (access & GENERIC_EXECUTE) access |= FILE_GENERIC_EXECUTE;
    if (access & GENERIC_ALL)     access |= FILE_ALL_ACCESS;
    return access & ~(GENERIC_READ | GENERIC_WRITE | GENERIC_EXECUTE | GENERIC_ALL);
}

thread_id_t get_thread_id (PTHREADINFO Thread)
{
    return (thread_id_t)PsGetThreadId(Thread->peThread);
}

process_id_t get_process_id(PPROCESSINFO Process)
{
    return (process_id_t)PsGetProcessId(Process->peProcess);
}

void set_fd_events( struct fd *fd, int events )
{
    UNIMPLEMENTED;
}

int check_fd_events( struct fd *fd, int events )
{
    UNIMPLEMENTED;
    return 0;
}

/* add a thread to an object wait queue; return 1 if OK, 0 on error */
int add_queue( struct object *obj, struct wait_queue_entry *entry )
{
#if 0
    grab_object( obj );
    entry->obj = obj;
    list_add_tail( &obj->wait_queue, &entry->entry );
#else
    UNIMPLEMENTED;
#endif
    return 1;
}

/* remove a thread from an object wait queue */
void remove_queue( struct object *obj, struct wait_queue_entry *entry )
{
#if 0
    list_remove( &entry->entry );
    release_object( obj );
#else
    UNIMPLEMENTED;
#endif
}

void set_event( PKEVENT event )
{
    KeSetEvent(event, EVENT_INCREMENT, FALSE);
}

void reset_event( struct event *event )
{
    UNIMPLEMENTED;
}

/* Stupid thing to hack around missing export of memcmp from the kernel */
int memcmp(const void *s1, const void *s2, size_t n)
{
    if (n != 0) {
        const unsigned char *p1 = s1, *p2 = s2;
        do {
            if (*p1++ != *p2++)
	            return (*--p1 - *--p2);
        } while (--n != 0);
    }
    return 0;
}

PVOID
NTAPI
ExReallocPool(PVOID OldPtr, ULONG NewSize, ULONG OldSize)
{
    PVOID NewPtr = ExAllocatePool(PagedPool, NewSize);

    if (OldPtr)
    {
        RtlCopyMemory(NewPtr, OldPtr, (NewSize < OldSize) ? NewSize : OldSize);
        ExFreePool(OldPtr);
    }

    return NewPtr;
}

char * __cdecl GdStrDup(const char *src)
{
    char *newstr = ExAllocatePool(PagedPool, strlen(src));
    strcpy(newstr, src);
    return newstr;
}

void *GdCalloc(size_t num, size_t size)
{
    PVOID Buf;
    ULONG TotalSize = num * size;

    Buf = ExAllocatePool(PagedPool, TotalSize);
    RtlZeroMemory(Buf, TotalSize);

    return Buf;
}

#define DIFFTIME 0x19db1ded53e8000ULL

time_t
SystemTimeToUnixTime(const PLARGE_INTEGER SystemTime)
{
    ULARGE_INTEGER ULargeInt;

    ULargeInt.QuadPart = SystemTime->QuadPart;
    ULargeInt.QuadPart -= DIFFTIME;

    return ULargeInt.QuadPart / 10000000;
}

BOOL NTAPI
RosUserRegisterShellHookWindow(HWND hWnd)
{
    //PPROCESSINFO process = PsGetCurrentProcessWin32Process();
    //struct desktop *desktop;
    struct list *entry, *shell_hooks;
    PSHELL_HOOK_WINDOW hook_entry;

    DPRINT("UserRegisterShellHookWindow\n");

    UserEnterExclusive();

    //if (process->desktop && (desktop = get_desktop_obj( process, process->desktop, 0 )))
    {
        shell_hooks = &global_shell_hooks;

        /* First deregister the window, so we can be sure it's never twice in the
        * list.
        */
        LIST_FOR_EACH(entry, shell_hooks)
        {
            hook_entry = LIST_ENTRY( entry, SHELL_HOOK_WINDOW, ListEntry );

            if (hook_entry->hWnd == hWnd)
            {
                list_remove(&hook_entry->ListEntry);
                ExFreePool(hook_entry);
                //release_object( desktop );
                break;
            }
        }

        hook_entry = ExAllocatePool(PagedPool, sizeof(SHELL_HOOK_WINDOW));

        if (!hook_entry)
        {
            //release_object( desktop );
            UserLeave();
            return FALSE;
        }

        hook_entry->hWnd = hWnd;
        list_add_tail(shell_hooks, &hook_entry->ListEntry);

        //release_object( desktop );
    }

    UserLeave();

    return TRUE;
}

BOOL NTAPI
RosUserDeRegisterShellHookWindow(HWND hWnd)
{
    //PPROCESSINFO process = PsGetCurrentProcessWin32Process();
    //struct desktop *desktop;
    struct list *entry, *shell_hooks;
    PSHELL_HOOK_WINDOW hook_entry;

    UserEnterExclusive();

    //if (process->desktop && (desktop = get_desktop_obj( process, process->desktop, 0 )))
    {
        shell_hooks = &global_shell_hooks;

        LIST_FOR_EACH(entry, shell_hooks)
        {
            hook_entry = LIST_ENTRY( entry, SHELL_HOOK_WINDOW, ListEntry );

            if (hook_entry->hWnd == hWnd)
            {
                list_remove(&hook_entry->ListEntry);
                ExFreePool(hook_entry);
                //release_object( desktop );
                UserLeave();
                return TRUE;
            }
        }

        //release_object( desktop );
    }

    UserLeave();
    return FALSE;
}

BOOL NTAPI
RosUserBuildShellHookHwndList(HWND *list, UINT *cbSize)
{
    //PPROCESSINFO process = PsGetCurrentProcessWin32Process();
    //struct desktop *desktop;
    struct list *entry, *shell_hooks;
    PSHELL_HOOK_WINDOW hook_entry;
    ULONG entries=0;
    NTSTATUS Status = STATUS_SUCCESS;

    UserEnterExclusive();

    //if (process->desktop && (desktop = get_desktop_obj( process, process->desktop, 0 )))
    {
        shell_hooks = &global_shell_hooks;

        /* fixme: if we save nb elements in desktop, we dont have to loop to find nb entries */
        LIST_FOR_EACH(entry, shell_hooks)
            entries++;

        if (!entries)
        {
            //release_object( desktop );
            UserLeave();
            return FALSE;
        }

        /* Check if enough user buffer was provided */
        if (*cbSize < sizeof(HWND) * entries)
        {
            *cbSize = sizeof(HWND) * entries;
            //release_object( desktop );
            UserLeave();
            return FALSE;
        }

        /* Fill the list */
        _SEH2_TRY
        {
            if (list)
            {
                HWND* cursor = list;

                LIST_FOR_EACH(entry, shell_hooks)
                {
                    hook_entry = LIST_ENTRY( entry, SHELL_HOOK_WINDOW, ListEntry );

                    *cursor++ = hook_entry->hWnd;
                }

                *cbSize = sizeof(HWND) * entries;
            }
        }
        _SEH2_EXCEPT(EXCEPTION_EXECUTE_HANDLER)
        {
            Status = _SEH2_GetExceptionCode();
        }
        _SEH2_END;

        /* Return error in case of exception */
        if (!NT_SUCCESS(Status))
        {
            //release_object( desktop );
            UserLeave();
            return FALSE;
        }

        //release_object( desktop );
    }

    UserLeave();

    return TRUE;
}

/* EOF */
